#include "tmEventSetup/esObjectHandle.hh"

#include "tmEventSetup/tmEventSetup.hh"
#include "tmEventSetup/esCutHandle.hh"
#include "tmGrammar/Cut.hh"
#include "tmUtil/tmUtil.hh"

#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/replace.hpp>

namespace tmeventsetup
{

// keys for tables
const std::string kName("name");
const std::string kObject("object");
const std::string kType("type");
const std::string kMinimum("minimum");
const std::string kMaximum("maximum");
const std::string kData("data");
const std::string kSeparator("-");
const std::string kSuffixSeparator("_");
const std::string kValueFormat("%+23.16E");

const std::map<std::string, esObjectType> ObjectTypeMap {
  {Object::MU, Muon},
  {Object::EG, Egamma},
  {Object::TAU, Tau},
  {Object::JET, Jet},
  {Object::ETT, ETT},
  {Object::HTT, HTT},
  {Object::ETM, ETM},
  {Object::HTM, HTM},
  {Object::MBT0HFP, MBT0HFP},
  {Object::MBT1HFP, MBT1HFP},
  {Object::MBT0HFM, MBT0HFM},
  {Object::MBT1HFM, MBT1HFM},
  {Object::TOWERCOUNT, TOWERCOUNT},
  {Object::ASYMET, ASYMET},
  {Object::ASYMHT, ASYMHT},
  {Object::ASYMETHF, ASYMETHF},
  {Object::ASYMHTHF, ASYMHTHF},
  {Object::CENT0, CENT0},
  {Object::CENT1, CENT1},
  {Object::CENT2, CENT2},
  {Object::CENT3, CENT3},
  {Object::CENT4, CENT4},
  {Object::CENT5, CENT5},
  {Object::CENT6, CENT6},
  {Object::CENT7, CENT7},
  {Object::MUS0, MUS0},
  {Object::MUS1, MUS1},
  {Object::MUSOOT0, MUSOOT0},
  {Object::MUSOOT1, MUSOOT1},
  {Object::ETTEM, ETTEM},
  {Object::ETMHF, ETMHF},
  {Object::EXT, EXT}
};

const std::map<esObjectType, std::string> ThresholdTypeMap {
  {Muon, ET_THR},
  {Egamma, ET_THR},
  {Tau, ET_THR},
  {Jet, ET_THR},
  {ETT, ET_THR},
  {HTT, ET_THR},
  {ETM, ET_THR},
  {HTM, ET_THR},
  {ETTEM, ET_THR},
  {ETMHF, ET_THR},
  {EXT, ET_THR},
  {MBT0HFP, COUNT},
  {MBT1HFP, COUNT},
  {MBT0HFM, COUNT},
  {MBT1HFM, COUNT},
  {TOWERCOUNT, COUNT},
  {ASYMET, COUNT},
  {ASYMHT, COUNT},
  {ASYMETHF, COUNT},
  {ASYMHTHF, COUNT},
  {CENT0, COUNT},
  {CENT1, COUNT},
  {CENT2, COUNT},
  {CENT3, COUNT},
  {CENT4, COUNT},
  {CENT5, COUNT},
  {CENT6, COUNT},
  {CENT7, COUNT},
  {MUS0, COUNT},
  {MUS1, COUNT},
  {MUSOOT0, COUNT},
  {MUSOOT1, COUNT}
};


template<typename T>
std::string formatValue(const T& value)
{
  return boost::str(boost::format(kValueFormat) % value);
}


esObjectHandle::esObjectHandle(const Object::Item& item)
{
  init(item);
}


esObjectHandle::esObjectHandle(const Object::Item& item, const tmtable::Table& cuts_in_algo)
{
  init(item);
  for (const auto& item_cut: item.cuts)
  {
    for (const auto& algo_cut: cuts_in_algo)
    {
      tmtable::Row& cut = const_cast<tmtable::Row&>(algo_cut);
      if (item_cut != cut[kName]) continue;

      // add object name if it is not included in the cut["type"]
      if (cut[kType].find(kSeparator) == std::string::npos)
      {
        cut[kType] = item.name + kSeparator + cut[kType];
      }
      addCut(esCutHandle(cut));
    }
  }
}


void
esObjectHandle::init(const Object::Item& item)
{
  name_ = item.getObjectName();

  type_ = static_cast<esObjectType>(Undef);

  std::string token = item.name;

  // TODO
  if (token.rfind(Object::EXT, 0) == 0)
  {
    token = Object::EXT;
  }

  const auto cit = ObjectTypeMap.find(token);
  if (cit == std::end(ObjectTypeMap))
  {
    TM_FATAL_ERROR("unknown type: " << TM_QUOTE(item.name));
  }
  type_ = cit->second;

  // If object is a signal
  if (Object::signalName.count(item.name))
  {
    bx_offset_ = boost::lexical_cast<int>(item.bx_offset);
    return;
  }

  if (type_ == EXT)
  {
    bx_offset_ = boost::lexical_cast<int>(item.bx_offset);
    ext_signal_name_ = getExternalName(name_);
    return;
  }

  comparison_operator_ = static_cast<esComparisonOperator>(Undef);
  if (item.comparison == Object::GE) comparison_operator_ = GE;
  else if (item.comparison == Object::NE) comparison_operator_ = NE;
  else if (item.comparison == Object::EQ) comparison_operator_ = EQ;
  else
  {
    TM_FATAL_ERROR("unknown comparison operator " << TM_QUOTE(item.comparison));
  }

  bx_offset_ = boost::lexical_cast<int>(item.bx_offset);
  threshold_ = Cut::parseThreshold(item.threshold);
  // Set proper cut type for object.
  const auto& cutType = getThresholdType();

  // Create table for threshold cut.
  tmtable::Row cut;
  cut[kName] = item.name + kSeparator + cutType + kSuffixSeparator + item.threshold; // MU-ET_1p2
  cut[kObject] = item.name;
  cut[kType] = item.name + kSeparator + cutType;
  cut[kMinimum] = formatValue(threshold_);
  cut[kMaximum] = "";
  cut[kData] = "";

  // Create threshold cut from table.
  addCut(esCutHandle(cut));
}


void
esObjectHandle::setExternalChannelId(const std::map<std::string, unsigned int>& map)
{
  if (type_ == EXT)
  {
    if (map.count(ext_signal_name_))
    {
      ext_channel_id_ = map.at(ext_signal_name_);
    }
    else
    {
      TM_FATAL_ERROR("unknown external signal name " << TM_QUOTE(ext_signal_name_));
    }
  }
}


void
esObjectHandle::addCut(const esCut& cut)
{
  cuts_.emplace_back(cut);
}


const std::string&
esObjectHandle::getThresholdType() const
{
  const auto key = static_cast<esObjectType>(type_);
  const auto cit = ThresholdTypeMap.find(key);
  if (cit == std::end(ThresholdTypeMap))
  {
    TM_FATAL_ERROR("unknown type: " << TM_QUOTE(type_));
  }
  return cit->second;
}

} // namespace tmeventsetup
