#include "tmEventSetup/esScaleHandle.hh"
#include "tmGrammar/Cut.hh"
#include "tmGrammar/Object.hh"
#include "tmUtil/tmUtil.hh"

#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include <algorithm>
#include <map>

namespace tmeventsetup
{

// keys for tables
const std::string kObject("object");
const std::string kType("type");
const std::string kMinimum("minimum");
const std::string kMaximum("maximum");
const std::string kStep("step");
const std::string kNBits("n_bits");

const std::map<std::string, esObjectType> ObjectTypeMap {
  {Object::MU, Muon},
  {Object::EG, Egamma},
  {Object::TAU, Tau},
  {Object::JET, Jet},
  {Object::HTT, HTT},
  {Object::ETT, ETT},
  {Object::ETM, ETM},
  {Object::HTM, HTM},
  {Object::MBT0HFP, MBT0HFP},
  {Object::MBT1HFP, MBT1HFP},
  {Object::MBT0HFM, MBT0HFM},
  {Object::MBT1HFM, MBT1HFM},
  {Object::ETTEM, ETTEM},
  {Object::ETMHF, ETMHF},
  {Object::TOWERCOUNT, TOWERCOUNT},
  {Object::ASYMET, ASYMET},
  {Object::ASYMHT, ASYMHT},
  {Object::ASYMETHF, ASYMETHF},
  {Object::ASYMHTHF, ASYMHTHF},
  {PRECISION, Precision}
};

const std::map<std::string, esScaleType> ScaleTypeMap {
  {ET_THR, EtScale},
  {Cut::UPT, UnconstrainedPtScale},
  {Cut::ETA, EtaScale},
  {Cut::PHI, PhiScale},
  {COUNT, CountScale},
  {PRECISION_DELTA, DeltaPrecision},
  {PRECISION_OVRM_DELTA, OvRmDeltaPrecision},
  {PRECISION_MASS, MassPrecision},
  {PRECISION_MASSPT, MassPtPrecision},
  {PRECISION_MATH, MathPrecision},
  {PRECISION_TBPT, TwoBodyPtPrecision},
  {PRECISION_TBPT_MATH, TwoBodyPtMathPrecision},
  {PRECISION_INVERSE_DR, InverseDeltaRPrecision}
};


/** Return search key for scale types mapping.
 * eg. "MU-MU-Math" -> "Math"
 */
std::string
getScaleTypeKey(const std::string& type)
{
  const auto pos = type.find_last_of("-");
  if (pos != std::string::npos)
  {
    return type.substr(pos + 1);
  }
  return type;
}


esScaleHandle::esScaleHandle(const tmtable::Row& scale)
{
  const auto& object = scale.at(kObject);
  const auto objectIt = ObjectTypeMap.find(object);
  if (objectIt == std::end(ObjectTypeMap))
  {
    TM_FATAL_ERROR("unknown object: " << TM_QUOTE(object));
  }
  object_ = objectIt->second;

  const auto& type = scale.at(kType);
  const auto typeIt = ScaleTypeMap.find(getScaleTypeKey(type));
  if (typeIt == std::end(ScaleTypeMap))
  {
    TM_FATAL_ERROR("unknown type: " << TM_QUOTE(type));
  }
  type_ = typeIt->second;

  name_ =  object + "-" + type;
  minimum_ = boost::lexical_cast<double>(scale.at(kMinimum));
  maximum_ = boost::lexical_cast<double>(scale.at(kMaximum));
  step_ = boost::lexical_cast<double>(scale.at(kStep));
  n_bits_ = boost::lexical_cast<unsigned int>(scale.at(kNBits));
}


void
esScaleHandle::addBin(const esBin& bin)
{
  bins_.emplace_back(bin);
}


void
esScaleHandle::sortBins()
{
  switch (type_)
  {
    // Sort eta scales by minimum to preserve two's complement hw_index
    case EtaScale:
      sortBinsByMinimum();
      break;
    default:
      sortBinsByHwIndex();
      break;
  }
}


void
esScaleHandle::sortBinsByMinimum()
{
  std::sort(std::begin(bins_), std::end(bins_), [](const esBin& a, const esBin& b) {
    return (a.minimum < b.minimum);
  });
}


void
esScaleHandle::sortBinsByHwIndex()
{
  std::sort(std::begin(bins_), std::end(bins_), [](const esBin& a, const esBin& b) {
    return (a.hw_index < b.hw_index);
  });
}

} // namespace tmeventsetup
