#include "tmEventSetup/esTriggerMenuHandle.hh"

#include "tmEventSetup/esBin.hh"
#include "tmEventSetup/esScaleHandle.hh"
#include "tmEventSetup/esCutHandle.hh"
#include "tmEventSetup/esObjectHandle.hh"
#include "tmEventSetup/esConditionHandle.hh"
#include "tmEventSetup/esAlgorithmHandle.hh"
#include "tmEventSetup/tmEventSetup.hh"

#include "tmGrammar/Cut.hh"
#include "tmGrammar/Function.hh"
#include "tmGrammar/Algorithm.hh"

#include "tmUtil/tmUtil.hh"

#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>

#include <cmath>
#include <cstdio>
#include <locale>
#include <utility>
#include <sstream>

namespace tmeventsetup
{

// keys for tables
const std::string kName("name");
const std::string kChannel("channel");
const std::string kNumber("number");
const std::string kMinimum("minimum");
const std::string kMaximum("maximum");
const std::string kValueFormat("%+23.16E");

// keys for conditions
const std::string kInstanceSeparator("_i");

// Retain old names for backward compatibility
const std::map<esConditionType, std::string> ConditionNameMap {
  {SingleMuon, "SingleMU"},
  {DoubleMuon, "DoubleMU"},
  {TripleMuon, "TripleMU"},
  {QuadMuon, "QuadMU"},
  {SingleEgamma, "SingleEG"},
  {DoubleEgamma, "DoubleEG"},
  {TripleEgamma, "TripleEG"},
  {QuadEgamma, "QuadEG"},
  {SingleTau, "SingleTAU"},
  {DoubleTau, "DoubleTAU"},
  {TripleTau, "TripleTAU"},
  {QuadTau, "QuadTAU"},
  {SingleJet, "SingleJET"},
  {DoubleJet, "DoubleJET"},
  {TripleJet, "TripleJET"},
  {QuadJet, "QuadJET"},
  {TotalEt, "SingleETT"},
  {TotalHt, "SingleHTT"},
  {MissingEt, "SingleETM"},
  {MissingHt, "SingleHTM"},
  {Externals, "SingleEXT"},
  {MuonMuonCorrelation, "MuonMuonCorrelation"},
  {MuonEsumCorrelation, "MuonEsumCorrelation"},
  {CaloMuonCorrelation, "CaloMuonCorrelation"},
  {CaloCaloCorrelation, "CaloCaloCorrelation"},
  {CaloEsumCorrelation, "CaloEsumCorrelation"},
  {InvariantMass, "InvariantMass"},
  {MinBiasHFP0, "SingleMBT0HFP"},
  {MinBiasHFP1, "SingleMBT1HFP"},
  {MinBiasHFM0, "SingleMBT0HFM"},
  {MinBiasHFM1, "SingleMBT1HFM"},
  {TotalEtEM, "SingleETEM"},
  {MissingEtHF, "SingleETMHF"},
  {TowerCount, "SingleTOWERCOUNT"},
  {TransverseMass, "TransverseMass"},
  {SingleEgammaOvRm, "SingleEgammaOvRm"},
  {DoubleEgammaOvRm, "DoubleEgammaOvRm"},
  {TripleEgammaOvRm, "TripleEgammaOvRm"},
  {QuadEgammaOvRm, "QuadEgammaOvRm"},
  {SingleTauOvRm, "SingleTauOvRm"},
  {DoubleTauOvRm, "DoubleTauOvRm"},
  {TripleTauOvRm, "TripleTauOvRm"},
  {QuadTauOvRm, "QuadTauOvRm"},
  {SingleJetOvRm, "SingleJetOvRm"},
  {DoubleJetOvRm, "DoubleJetOvRm"},
  {TripleJetOvRm, "TripleJetOvRm"},
  {QuadJetOvRm, "QuadJetOvRm"},
  {CaloCaloCorrelationOvRm, "CaloCaloCorrelationOvRm"},
  {InvariantMassOvRm, "InvariantMassOvRm"},
  {TransverseMassOvRm, "TransverseMassOvRm"},
  {AsymmetryEt, "SingleASYMET"},
  {AsymmetryHt, "SingleASYMHT"},
  {AsymmetryEtHF, "SingleASYMETHF"},
  {AsymmetryHtHF, "SingleASYMHTHF"},
  {Centrality0, "SingleCENT0"},
  {Centrality1, "SingleCENT1"},
  {Centrality2, "SingleCENT2"},
  {Centrality3, "SingleCENT3"},
  {Centrality4, "SingleCENT4"},
  {Centrality5, "SingleCENT5"},
  {Centrality6, "SingleCENT6"},
  {Centrality7, "SingleCENT7"},
  {InvariantMass3, "InvariantMass3"},
  {InvariantMassDeltaR, "InvariantMassDeltaR"},
  {InvariantMassUpt, "InvariantMassUpt"},
  {MuonShower0, "MuonShower0"},
  {MuonShower1, "MuonShower1"},
  {MuonShowerOutOfTime0, "MuonShowerOutOfTime0"},
  {MuonShowerOutOfTime1, "MuonShowerOutOfTime1"}
};


template<typename T>
std::string formatValue(const T& value)
{
  return boost::str(boost::format(kValueFormat) % value);
}


/** Returns const refrence to scale table or throws a std::runtime_error if
 * the required scale does not exist.
 */
const tmtable::Table& getScaleTable(const tmtable::StringTableMap& bins, const std::string& key)
{
    const auto rc = bins.find(key);

    if (rc == std::end(bins))
    {
      TM_FATAL_ERROR("missing scale set: " << TM_QUOTE(key));
    }

    return rc->second;
}


std::vector<std::string>
esTriggerMenuHandle::parse(const std::string& algorithm)
{
  Algorithm::Logic::clear();
  if (not Algorithm::parser(algorithm))
  {
    TM_FATAL_ERROR("parser error: " << TM_QUOTE(algorithm));
  }

  return Algorithm::Logic::tokens;
}


const std::string&
esTriggerMenuHandle::getConditionName(const int type) const
{
  const auto key = static_cast<esConditionType>(type);
  const auto it = ConditionNameMap.find(key);

  if (it == std::end(ConditionNameMap))
  {
    TM_FATAL_ERROR("unknown condition type: " << TM_QUOTE(type));
  }

  return it->second;
}


esCondition
esTriggerMenuHandle::getObjectCondition(const std::string& token,
                                        const tmtable::Table& cuts_in_algo) const
{
  Object::Item item;
  Object::parser(token, item);
  esObjectHandle objectHandle(item, cuts_in_algo);
  objectHandle.setExternalChannelId(external_map_);
  esObject& object = objectHandle;

  esConditionHandle conditionHandle;
  conditionHandle.addObject(object);

  switch (objectHandle.getType())
  {
    case Muon: conditionHandle.setType(SingleMuon); break;
    case Egamma: conditionHandle.setType(SingleEgamma); break;
    case Tau: conditionHandle.setType(SingleTau); break;
    case Jet: conditionHandle.setType(SingleJet); break;
    case ETT: conditionHandle.setType(TotalEt); break;
    case HTT: conditionHandle.setType(TotalHt); break;
    case ETM: conditionHandle.setType(MissingEt); break;
    case HTM: conditionHandle.setType(MissingHt); break;
    case EXT: conditionHandle.setType(Externals); break;
    case MBT0HFP: conditionHandle.setType(MinBiasHFP0); break;
    case MBT1HFP: conditionHandle.setType(MinBiasHFP1); break;
    case MBT0HFM: conditionHandle.setType(MinBiasHFM0); break;
    case MBT1HFM: conditionHandle.setType(MinBiasHFM1); break;
    case ETTEM: conditionHandle.setType(TotalEtEM); break;
    case ETMHF: conditionHandle.setType(MissingEtHF); break;
    case TOWERCOUNT: conditionHandle.setType(TowerCount); break;
    case ASYMET: conditionHandle.setType(AsymmetryEt); break;
    case ASYMHT: conditionHandle.setType(AsymmetryHt); break;
    case ASYMETHF: conditionHandle.setType(AsymmetryEtHF); break;
    case ASYMHTHF: conditionHandle.setType(AsymmetryHtHF); break;
    case CENT0: conditionHandle.setType(Centrality0); break;
    case CENT1: conditionHandle.setType(Centrality1); break;
    case CENT2: conditionHandle.setType(Centrality2); break;
    case CENT3: conditionHandle.setType(Centrality3); break;
    case CENT4: conditionHandle.setType(Centrality4); break;
    case CENT5: conditionHandle.setType(Centrality5); break;
    case CENT6: conditionHandle.setType(Centrality6); break;
    case CENT7: conditionHandle.setType(Centrality7); break;
    case MUS0: conditionHandle.setType(MuonShower0); break;
    case MUS1: conditionHandle.setType(MuonShower1); break;
    case MUSOOT0: conditionHandle.setType(MuonShowerOutOfTime0); break;
    case MUSOOT1: conditionHandle.setType(MuonShowerOutOfTime1); break;
    default:
      TM_FATAL_ERROR("not implemented: " << objectHandle.getType());
      break;
  }



  // If token already occurred set instance name, else create new instance name.
  if (token_to_condition_.count(token))
  {
    conditionHandle.setName(token_to_condition_.at(token));
  }
  else
  {
    std::ostringstream name;
    name << getConditionName(conditionHandle.getType()) << kInstanceSeparator << token_to_condition_.size(); // enumerate
    conditionHandle.setName(name.str());
  }

  return conditionHandle;
}


esCondition
esTriggerMenuHandle::getFunctionCondition(const std::string& token,
                                          const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  esCondition condition;
  Function::Item item;
  Function::parser(token, item);

  if (item.type == Function::Combination)
  {
    condition = getCombCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::CombinationOvRm)
  {
    condition = getOverlapRemovalCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::Distance)
  {
    condition = getDistCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::DistanceOvRm)
  {
    condition = getDistOverlapRemovalCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::InvariantMass)
  {
    condition = getMassCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::InvariantMassUpt)
  {
    condition = getMassUptCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::InvariantMassDeltaR)
  {
    condition = getMassDeltaRCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::InvariantMass3)
  {
    condition = getMassCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::InvariantMassOvRm)
  {
    condition = getMassOverlapRemovalCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::TransverseMass)
  {
    condition = getMassCondition(item, cuts_in_algo);
  }
  else if (item.type == Function::TransverseMassOvRm)
  {
    condition = getMassOverlapRemovalCondition(item, cuts_in_algo);
  }
  else
  {
    TM_FATAL_ERROR("unknown token: " << TM_QUOTE(token));
  }

  // If token already occurred set instance name, else create new instance name.
  if (token_to_condition_.count(token))
  {
    condition.setName(token_to_condition_.at(token));
  }
  else
  {
    std::ostringstream name;
    name << condition.getName() << kInstanceSeparator << token_to_condition_.size(); // enumerate
    condition.setName(name.str());
  }

  return condition;
}


esCondition
esTriggerMenuHandle::getCombCondition(const Function::Item& item,
                                      const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  // multi-object condition
  esConditionHandle conditionHandle;

  // populate condition with objects
  for (const auto& object: item.objects)
  {
    conditionHandle.addObject(esObjectHandle(object, cuts_in_algo));
  }

  // populate condition with cuts
  for (const auto& cut: item.cuts)
  {
    for (const auto& row: cuts_in_algo)
    {
      if (cut == row.at(kName))
      {
        conditionHandle.addCut(esCutHandle(row));
      }
    }
  }

  const esObjectType type = static_cast<esObjectType>(conditionHandle.getObjects().at(0).getType());

  // count of objects
  const size_t count = conditionHandle.getObjects().size();

  // distinguish number of objects
  switch (count)
  {
    case 2: // double condition
      switch (type)
      {
        case Muon: conditionHandle.setType(DoubleMuon); break;
        case Egamma: conditionHandle.setType(DoubleEgamma); break;
        case Tau: conditionHandle.setType(DoubleTau); break;
        case Jet: conditionHandle.setType(DoubleJet); break;
        default:
          TM_FATAL_ERROR("not implemented: " << TM_QUOTE(type));
          break;
      }
      break;

    case 3: // triple condition
      switch (type)
      {
        case Muon: conditionHandle.setType(TripleMuon); break;
        case Egamma: conditionHandle.setType(TripleEgamma); break;
        case Tau: conditionHandle.setType(TripleTau); break;
        case Jet: conditionHandle.setType(TripleJet); break;
        default:
          TM_FATAL_ERROR("not implemented: " << TM_QUOTE(type));
          break;
      }
      break;

    case 4: // quad condition
      switch (type)
      {
        case Muon: conditionHandle.setType(QuadMuon); break;
        case Egamma: conditionHandle.setType(QuadEgamma); break;
        case Tau: conditionHandle.setType(QuadTau); break;
        case Jet: conditionHandle.setType(QuadJet); break;
        default:
          TM_FATAL_ERROR("not implemented: " << TM_QUOTE(type));
          break;
      }
      break;

    default:
      TM_FATAL_ERROR("unsupported # of objects: " << count);
      break;
  }

  // Set condition name
  const auto& name = getConditionName(conditionHandle.getType());
  conditionHandle.setName(name);

  return conditionHandle;
}


esCondition
esTriggerMenuHandle::getOverlapRemovalCondition(const Function::Item& item,
                                                const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  esConditionHandle conditionHandle;

  // populate condition with objects
  for (const auto& object: item.objects)
  {
    conditionHandle.addObject(esObjectHandle(object, cuts_in_algo));
  }

  // populate condition with cuts
  for (const auto& cut: item.cuts)
  {
    for (const auto& row: cuts_in_algo)
    {
      if (cut == row.at(kName))
      {
        conditionHandle.addCut(esCutHandle(row));
      }
    }
  }

  // get list of objects to process
  const std::vector<esObject>& objects = conditionHandle.getObjects();

  // get type of first object
  const esObjectType type = static_cast<esObjectType>(objects.front().getType());

  // get type of possible different object (always last one)
  const esObjectType typeReference = static_cast<esObjectType>(objects.back().getType());

  // check for appended reference object of different type
  const bool hasReference = (type != typeReference);

  // count of objects (without possible reference)
  const size_t count = (hasReference ? objects.size() - 1 : objects.size());

  // distinguish number of left side objects
  switch (count)
  {
    case 1: // SingleOverlapRemoval
      switch (type)
      {
        case Egamma: conditionHandle.setType(SingleEgammaOvRm); break;
        case Tau: conditionHandle.setType(SingleTauOvRm); break;
        case Jet: conditionHandle.setType(SingleJetOvRm); break;
        default:
          TM_FATAL_ERROR("not implemented: " << TM_QUOTE(type));
          break;
      }
      break;
    case 2: // DoubleOverlapRemoval
      switch (type)
      {
        case Egamma: conditionHandle.setType(DoubleEgammaOvRm); break;
        case Tau: conditionHandle.setType(DoubleTauOvRm); break;
        case Jet: conditionHandle.setType(DoubleJetOvRm); break;
        default:
          TM_FATAL_ERROR("not implemented: " << TM_QUOTE(type));
          break;
      }
      break;

    case 3: // TripleOverlapRemoval
      switch (type)
      {
        case Egamma: conditionHandle.setType(TripleEgammaOvRm); break;
        case Tau: conditionHandle.setType(TripleTauOvRm); break;
        case Jet: conditionHandle.setType(TripleJetOvRm); break;
        default:
          TM_FATAL_ERROR("not implemented: " << TM_QUOTE(type));
          break;
      }
      break;

    case 4: // QuadOverlapRemoval
      switch (type)
      {
        case Egamma: conditionHandle.setType(QuadEgammaOvRm); break;
        case Tau: conditionHandle.setType(QuadTauOvRm); break;
        case Jet: conditionHandle.setType(QuadJetOvRm); break;
        default:
          TM_FATAL_ERROR("not implemented: " << TM_QUOTE(type));
          break;
      }
      break;

    default:
      TM_FATAL_ERROR("unsupported # of objects: " << count);
      break;
  }

  // Set condition name
  const auto& name = getConditionName(conditionHandle.getType());
  conditionHandle.setName(name);

  return conditionHandle;
}


esCondition
esTriggerMenuHandle::getDistCondition(const Function::Item& item,
                                      const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  // distance (correlation) condition
  if (item.objects.size() != 2)
  {
    TM_FATAL_ERROR("# of objects != 2: " << item.objects.size());
  }

  esConditionHandle conditionHandle;
  for (const auto& object: item.objects)
  {
    conditionHandle.addObject(esObjectHandle(object, cuts_in_algo));
  }

  size_t nCut = item.cuts.size();
  if (nCut == 0 or nCut > 3)  // deta/dphi/chgcor combination possible
  {
    TM_FATAL_ERROR("# of cuts not in range [1,3]: " << nCut);
  }

  bool hasDistCut = false;
  for (const auto& row: cuts_in_algo)
  {
    for (const auto& cut: item.cuts)
    {
      if (cut == row.at(kName))
      {
        const esCutHandle cutHandle(row);
        switch (cutHandle.getCutType())
        {
          case DeltaEta:
          case DeltaPhi:
          case DeltaR:
            hasDistCut = true;
            break;
          case ChargeCorrelation:
            break;
          case TwoBodyPt:
            break;
          default:
            TM_FATAL_ERROR("unknown cut type: " << cutHandle.getCutType());
        }
        conditionHandle.addCut(std::move(cutHandle));
      }
    }
  }

  if (not hasDistCut)
  {
    TM_FATAL_ERROR("no dist cut specified");
  }


  const esObjectType type1 = static_cast<esObjectType>(conditionHandle.getObjects().at(0).getType());
  const esObjectType type2 = static_cast<esObjectType>(conditionHandle.getObjects().at(1).getType());
  const esCombinationType combination = getObjectCombination(type1, type2);

  switch (combination)
  {
    case MuonMuonCombination:
      conditionHandle.setType(MuonMuonCorrelation);
      break;

    case MuonEsumCombination:
      conditionHandle.setType(MuonEsumCorrelation);
      break;

    case CaloMuonCombination:
      conditionHandle.setType(CaloMuonCorrelation);
      break;

    case CaloCaloCombination:
      conditionHandle.setType(CaloCaloCorrelation);
      break;

    case CaloEsumCombination:
      conditionHandle.setType(CaloEsumCorrelation);
      break;

    default:
      TM_FATAL_ERROR("unknown combination type: " << combination);
  }

  // Set condition name
  const auto& name = getConditionName(conditionHandle.getType());
  conditionHandle.setName(name);

  return conditionHandle;
}


esCondition
esTriggerMenuHandle::getDistOverlapRemovalCondition(const Function::Item& item,
                                                    const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  // distance (correlation) condition with overlap removal
  if (item.objects.size() < 2 or
      item.objects.size() > 3)
  {
    TM_FATAL_ERROR("# of objects != 2 or 2+1: " << item.objects.size());
  }

  esConditionHandle conditionHandle;
  for (const auto& object: item.objects)
  {
    conditionHandle.addObject(esObjectHandle(object, cuts_in_algo));
  }

  size_t nCut = item.cuts.size();
  if (nCut == 0 or nCut > 6)  // deta/dphi/chgcor/ormdeta/ormdphi/ormdr combination possible
  {
    TM_FATAL_ERROR("# of cuts not in range [1,6]: " << nCut);
  }

  bool hasDistCut = false;
  bool hasOvRmCut = false;
  for (const auto& row: cuts_in_algo)
  {
    for (const auto& cut: item.cuts)
    {
      if (cut == row.at(kName))
      {
        const esCutHandle cutHandle(row);
        switch (cutHandle.getCutType())
        {
          case DeltaEta:
          case DeltaPhi:
          case DeltaR:
            hasDistCut = true;
            break;
          case OvRmDeltaEta:
          case OvRmDeltaPhi:
          case OvRmDeltaR:
            hasOvRmCut = true;
            break;
          case ChargeCorrelation:
            break;
          case TwoBodyPt:
            break;
          default:
            TM_FATAL_ERROR("unknown cut type: " << cutHandle.getCutType());
        }
        conditionHandle.addCut(std::move(cutHandle));
      }
    }
  }

  if (not hasDistCut)
  {
    TM_FATAL_ERROR("missing dist cut (mandatory)");
  }

  if (not hasOvRmCut)
  {
    TM_FATAL_ERROR("missing overlap removal cut (mandatory)");
  }

  const esObjectType type1 = static_cast<esObjectType>(conditionHandle.getObjects().at(0).getType());
  const esObjectType type2 = static_cast<esObjectType>(conditionHandle.getObjects().at(1).getType());
  const esCombinationType combination = getObjectCombination(type1, type2);

  switch (combination)
  {
    case CaloCaloCombination: // NOTE !
      conditionHandle.setType(CaloCaloCorrelationOvRm); // NOTE !
      break;

    default:
      TM_FATAL_ERROR("unknown combination type: " << combination);
  }

  // Set condition name
  const auto& name = getConditionName(conditionHandle.getType());
  conditionHandle.setName(name);

  return conditionHandle;
}


esCondition
esTriggerMenuHandle::getMassCondition(const Function::Item& item,
                                      const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  // verify number of objects
  if (item.type == Function::InvariantMass3)
  {
    if (item.objects.size() != 3)
    {
      TM_FATAL_ERROR("# of objects != 3: " << item.objects.size());
    }
  }
  else
  {
    if (item.objects.size() != 2)
    {
      TM_FATAL_ERROR("# of objects != 2: " << item.objects.size());
    }
  }

  esConditionHandle conditionHandle;

  // assign objects
  for (const auto& object: item.objects)
  {
    conditionHandle.addObject(esObjectHandle(object, cuts_in_algo));
  }

  const size_t nCut = item.cuts.size();

  // TODO: ...
  if (item.type == Function::InvariantMass3)
  {
    if (nCut == 0 or nCut > 2)  // mass
    {
      TM_FATAL_ERROR("# of cuts not in range [1,2]: " << nCut);
    }
  }
  else
  {
    if (nCut == 0 or nCut > 5)  // mass/chgcor/dR|(dEta/dPhi)/tbpt
    {
      TM_FATAL_ERROR("# of cuts not in range [1,5]: " << nCut);
    }
  }

  bool hasMassCut = false;

  // assign function cuts
  for (const auto& row: cuts_in_algo)
  {
    for (const auto& cut: item.cuts)
    {
      if (cut == row.at(kName))
      {
        const esCutHandle cutHandle(row);
        if (cutHandle.getCutType() == Mass) hasMassCut = true;
        conditionHandle.addCut(std::move(cutHandle));
      }
    }
  }

  if (not hasMassCut)
  {
    TM_FATAL_ERROR("no MASS cut specified");
  }

  if (item.type == Function::InvariantMass)
    conditionHandle.setType(InvariantMass);
  else if (item.type == Function::InvariantMass3)
    conditionHandle.setType(InvariantMass3);
  else if (item.type == Function::TransverseMass)
    conditionHandle.setType(TransverseMass);

  // Set condition name
  const auto& name = getConditionName(conditionHandle.getType());
  conditionHandle.setName(name);

  return conditionHandle;
}


esCondition
esTriggerMenuHandle::getMassUptCondition(const Function::Item& item,
                                         const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  // verify number of objects
  if (item.objects.size() != 2)
  {
    TM_FATAL_ERROR("# of objects != 2: " << item.objects.size());
  }

  esConditionHandle conditionHandle;

  // assign objects
  for (const auto& object: item.objects)
  {
    conditionHandle.addObject(esObjectHandle(object, cuts_in_algo));
  }

  const size_t nCut = item.cuts.size();

  if (nCut == 0 or nCut > 5)  // massupt/chgcor/dR|(dEta/dPhi)/tbpt
  {
    TM_FATAL_ERROR("# of cuts not in range [1,5]: " << nCut);
  }

  bool hasMassUptCut = false;

  // assign function cuts
  for (const auto& row: cuts_in_algo)
  {
    for (const auto& cut: item.cuts)
    {
      if (cut == row.at(kName))
      {
        const esCutHandle cutHandle(row);
        if (cutHandle.getCutType() == MassUpt) hasMassUptCut = true;
        conditionHandle.addCut(std::move(cutHandle));
      }
    }
  }

  if (not hasMassUptCut)
  {
    TM_FATAL_ERROR("no MASSUPT cut specified");
  }

  conditionHandle.setType(InvariantMassUpt);

  // Set condition name
  const auto& name = getConditionName(conditionHandle.getType());
  conditionHandle.setName(name);

  return conditionHandle;
}


esCondition
esTriggerMenuHandle::getMassDeltaRCondition(const Function::Item& item,
                                            const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  // verify number of objects
  if (item.objects.size() != 2)
  {
    TM_FATAL_ERROR("# of objects != 2: " << item.objects.size());
  }

  esConditionHandle conditionHandle;

  // assign objects
  for (const auto& object: item.objects)
  {
    conditionHandle.addObject(esObjectHandle(object, cuts_in_algo));
  }

  size_t nCut = item.cuts.size();

  if (nCut == 0 or nCut > 5)  // mass/chgcor/dR|(dEta/dPhi)/tbpt
  {
    TM_FATAL_ERROR("# of cuts not in range [1,5]: " << nCut);
  }

  bool hasMassDeltaRCut = false;

  // assign function cuts
  for (const auto& row: cuts_in_algo)
  {
    for (const auto& cut: item.cuts)
    {
      if (cut == row.at(kName))
      {
        const esCutHandle cutHandle(row);
        if (cutHandle.getCutType() == MassDeltaR) hasMassDeltaRCut = true;
        conditionHandle.addCut(std::move(cutHandle));
      }
    }
  }

  if (not hasMassDeltaRCut)
  {
    TM_FATAL_ERROR("no MASSDR cut specified");
  }

  conditionHandle.setType(InvariantMassDeltaR);

  // Set condition name
  const auto& name = getConditionName(conditionHandle.getType());
  conditionHandle.setName(name);

  return conditionHandle;
}


esCondition
esTriggerMenuHandle::getMassOverlapRemovalCondition(const Function::Item& item,
                                                    const tmtable::Table& cuts_in_algo) const
{
  TM_LOG_DBG("-");

  // invariant-mass condition
  if (item.objects.size() < 2 or
      item.objects.size() > 3)
  {
    TM_FATAL_ERROR("# of objects != 2 or 2+1: " << item.objects.size());
  }

  esConditionHandle conditionHandle;
  for (const auto& object: item.objects)
  {
    conditionHandle.addObject(esObjectHandle(object, cuts_in_algo));
  }

  size_t nCut = item.cuts.size();
  if (nCut == 0 or nCut > 7)  // mass/dR|(dEta/dPhi)/tbpt
  {
    TM_FATAL_ERROR("# of cuts not in range [1,7]: " << nCut);
  }

  bool hasMassCut = false;
  bool hasOvRmCut = false;

  for (const auto& row: cuts_in_algo)
  {
    for (const auto& cut: item.cuts)
    {
      if (cut == row.at(kName))
      {
        const esCutHandle cutHandle(row);
        if (cutHandle.getCutType() == Mass) hasMassCut = true;
        if (cutHandle.getCutType() == OvRmDeltaEta) hasOvRmCut = true;
        if (cutHandle.getCutType() == OvRmDeltaPhi) hasOvRmCut = true;
        if (cutHandle.getCutType() == OvRmDeltaR) hasOvRmCut = true;
        conditionHandle.addCut(std::move(cutHandle));
      }
    }
  }

  if (not hasMassCut)
  {
    TM_FATAL_ERROR("missing mass cut (mandatory)");
  }

  if (not hasOvRmCut)
  {
    TM_FATAL_ERROR("missing overlap removal cut (mandatory)");
  }

  if (item.type == Function::InvariantMassOvRm)
    conditionHandle.setType(InvariantMassOvRm);
  else if (item.type == Function::TransverseMassOvRm)
    conditionHandle.setType(TransverseMassOvRm);

  // Set condition name
  const auto& name = getConditionName(conditionHandle.getType());
  conditionHandle.setName(name);

  return conditionHandle;
}


void
esTriggerMenuHandle::setConditionMap(const std::string& token,
                                     const tmtable::Table& cuts_in_algo)
{
  TM_LOG_DBG(TM_VALUE_DBG(condition_map_.size()));
  if (Algorithm::isGate(token)) return;

  esCondition condition;

  if (Object::isObject(token))
  {
    condition = getObjectCondition(token, cuts_in_algo);
  }
  else if (Function::isFunction(token))
  {
    condition = getFunctionCondition(token, cuts_in_algo);
  }
  else
  {
    TM_FATAL_ERROR("unknown token: " << TM_QUOTE(token));
  }

  std::pair<std::map<std::string, std::string>::iterator, bool> rc;
  rc = token_to_condition_.insert(std::make_pair(token, condition.getName()));

  condition_map_.insert(std::make_pair(condition.getName(), condition));
#if defined(SWIG)
  condition_map_p_[condition.getName()] = &(condition_map_[condition.getName()]);
#endif
}



void
esTriggerMenuHandle::setAlgorithmMap(const tmtable::Row& algorithm)
{
  TM_LOG_DBG(TM_VALUE_DBG(algorithm_map_.size()));

  esAlgorithmHandle algoHandle;
  algoHandle.init(algorithm);
  algoHandle.setExpressionInCondition(token_to_condition_);
  algoHandle.setRpnVector(parse(algoHandle.getExpressionInCondition()));

  esAlgorithm& algo = algoHandle;

  std::pair<std::map<std::string, esAlgorithm>::iterator, bool> rc;
  rc = algorithm_map_.insert(std::make_pair(algoHandle.getName(), algo));
  if (not rc.second)
  {
    TM_LOG_ERR("failed to insert algorithm: " << TM_QUOTE(algoHandle.getName()));
  }
#if defined(SWIG)
  algorithm_map_p_[algoHandle.getName()] = &(algorithm_map_[algoHandle.getName()]);
#endif
}


void
esTriggerMenuHandle::setScaleMap(const tmtable::Scale& scale)
{
  for (const auto& scales: scale.scales)
  {
    esScaleHandle scaleHandle(scales);
    if (scaleHandle.getObjectType() != Precision)
    {
      for (const auto& bin: getScaleTable(scale.bins, scaleHandle.getName()))
      {
        const int id =
          boost::lexical_cast<unsigned int>(bin.at(kNumber));
        const double min =
          boost::lexical_cast<double>(bin.at(kMinimum));
        const double max =
          boost::lexical_cast<double>(bin.at(kMaximum));
        scaleHandle.addBin(esBin(id, min, max));
      }
      scaleHandle.sortBins();
    }
    scale_map_[scaleHandle.getName()] = scaleHandle;
#if defined(SWIG)
    scale_map_p_[scaleHandle.getName()] = &(scale_map_[scaleHandle.getName()]);
#endif
  }
}


void
esTriggerMenuHandle::setExternalMap(const tmtable::ExtSignal& map)
{
  for (const auto& signal: map.extSignals)
  {
    const auto& name = signal.at(kName);
    const auto& value = signal.at(kChannel);
    external_map_[name] = boost::lexical_cast<unsigned int>(value);
    TM_LOG_DBG(TM_QUOTE(name) << " : " << TM_QUOTE(value));
  }
}


unsigned int
esTriggerMenuHandle::getIndex(const esCutValue& cut, const std::string& range, const tmtable::Table& bins) const
{
  // Convert double to string representation.
  const std::string real = formatValue(cut.value);
  TM_LOG_DBG(TM_VALUE_DBG(real));
  TM_LOG_DBG(TM_VALUE_DBG(range));

  unsigned int index = std::numeric_limits<unsigned int>::max();
  for (const auto& bin: bins)
  {
    // TODO: segfaults if key 'range' not found
    if (bin.at(range) == real)
    {
      // TODO: segfaults if key 'number' not found
      std::istringstream ss(bin.at(kNumber));
      if (not (ss >> index)) index = std::numeric_limits<unsigned int>::max();
      break;
    }
  }
  return index;
}



void
esTriggerMenuHandle::setHwIndex(const tmtable::StringTableMap& bins)
{
  TM_LOG_DBG("");
  for (std::map<std::string, esCondition>::const_iterator cit = std::begin(condition_map_);
       cit != std::end(condition_map_); cit++)
  {
    const std::vector<esObject>& objects = cit->second.getObjects();
    for (size_t ii = 0; ii < objects.size(); ii++)
    {
      const esObject& object = objects.at(ii);
      const std::vector<esCut>& cuts = object.getCuts();
      for (size_t jj = 0; jj < cuts.size(); jj++)
      {
        esCut& cut = const_cast<esCut&>(cuts.at(jj));
        if (cut.getObjectType() == static_cast<esObjectType>(Undef)) continue;
        const std::string key = cut.getKey();
        TM_LOG_DBG(TM_VALUE_DBG(key));

        const esCutType type = static_cast<esCutType>(cut.getCutType());
        if ((type == Threshold) or (type == Count))
        {
          const tmtable::Table& table = getScaleTable(bins, key);
          const esCutValue& cutValue = cut.getMinimum();
          cut.setMinimumIndex(getIndex(cutValue, kMinimum, table));
        }
        else if ((type == UnconstrainedPt) or (type == Eta) or (type == Phi))
        {
          const tmtable::Table& table = getScaleTable(bins, key);
          const esCutValue& minimum = cut.getMinimum();
          cut.setMinimumIndex(getIndex(minimum, kMinimum, table));
          const esCutValue& maximum = cut.getMaximum();
          cut.setMaximumIndex(getIndex(maximum, kMaximum, table));
        }
      }
    }
  }
}


void
esTriggerMenuHandle::setPrefix4Precision(const std::vector<esObject>& objects,
                                         std::string& prefix)
{
  // Allow more than two objects to accept overlap removal conditions (third object).
  if (objects.size() < 2)
  {
    TM_FATAL_ERROR("# of objects < 2: " << objects.size());
  }
  const esObject& o1 = objects.at(0); // first type
  const esObject& o2 = objects.at(1); // second type

  switch (o1.getType())
  {
    case Egamma:
      switch (o2.getType())
      {
        case Egamma:
          prefix = "PRECISION-EG-EG-";
          break;
        case Tau:
          prefix = "PRECISION-EG-TAU-";
          break;
        case Jet:
          prefix = "PRECISION-EG-JET-";
          break;
        case Muon:
          prefix = "PRECISION-EG-MU-";
          break;
        case ETM:
          prefix = "PRECISION-EG-ETM-";
          break;
        case HTM:
          prefix = "PRECISION-EG-HTM-";
          break;
        case ETMHF:
          prefix = "PRECISION-EG-ETMHF-";
          break;
        default:
          TM_FATAL_ERROR("unsupported type: " << o2.getType());
      }
      break;

    case Tau:
      switch (o2.getType())
      {
        case Egamma:
          prefix = "PRECISION-EG-TAU-";
          break;
        case Tau:
          prefix = "PRECISION-TAU-TAU-";
          break;
        case Jet:
          prefix = "PRECISION-JET-TAU-";
          break;
        case Muon:
          prefix = "PRECISION-TAU-MU-";
          break;
        case ETM:
          prefix = "PRECISION-TAU-ETM-";
          break;
        case HTM:
          prefix = "PRECISION-TAU-HTM-";
          break;
        case ETMHF:
          prefix = "PRECISION-TAU-ETMHF-";
          break;
        default:
          TM_FATAL_ERROR("unsupported type: " << o2.getType());
      }
      break;

    case Jet:
      switch (o2.getType())
      {
        case Egamma:
          prefix = "PRECISION-EG-JET-";
          break;
        case Tau:
          prefix = "PRECISION-JET-TAU-";
          break;
        case Jet:
          prefix = "PRECISION-JET-JET-";
          break;
        case Muon:
          prefix = "PRECISION-JET-MU-";
          break;
        case ETM:
          prefix = "PRECISION-JET-ETM-";
          break;
        case HTM:
          prefix = "PRECISION-JET-HTM-";
          break;
        case ETMHF:
          prefix = "PRECISION-JET-ETMHF-";
          break;
        default:
          TM_FATAL_ERROR("unsupported type: " << o2.getType());
      }
      break;

    case Muon:
      switch (o2.getType())
      {
        case Egamma:
          prefix = "PRECISION-EG-MU-";
          break;
        case Tau:
          prefix = "PRECISION-TAU-MU-";
          break;
        case Jet:
          prefix = "PRECISION-JET-MU-";
          break;
        case Muon:
          prefix = "PRECISION-MU-MU-";
          break;
        case ETM:
          prefix = "PRECISION-MU-ETM-";
          break;
        case HTM:
          prefix = "PRECISION-MU-HTM-";
          break;
        case ETMHF:
          prefix = "PRECISION-MU-ETMHF-";
          break;
        default:
          TM_FATAL_ERROR("unsupported type: " << o2.getType());
      }
      break;

    case ETM:
      switch (o2.getType())
      {
        case Egamma:
          prefix = "PRECISION-EG-ETM-";
          break;
        case Tau:
          prefix = "PRECISION-TAU-ETM-";
          break;
        case Jet:
          prefix = "PRECISION-JET-ETM-";
          break;
        case Muon:
          prefix = "PRECISION-MU-ETM-";
          break;
        default:
          TM_FATAL_ERROR("unsupported type: " << o2.getType());
      }
      break;

    case HTM:
      switch (o2.getType())
      {
        case Egamma:
          prefix = "PRECISION-EG-HTM-";
          break;
        case Tau:
          prefix = "PRECISION-TAU-HTM-";
          break;
        case Jet:
          prefix = "PRECISION-JET-HTM-";
          break;
        case Muon:
          prefix = "PRECISION-MU-HTM-";
          break;
        default:
          TM_FATAL_ERROR("unsupported type: " << o2.getType());
      }
      break;

    case ETMHF:
      switch (o2.getType())
      {
        case Egamma:
          prefix = "PRECISION-EG-ETMHF-";
          break;
        case Tau:
          prefix = "PRECISION-TAU-ETMHF-";
          break;
        case Jet:
          prefix = "PRECISION-JET-ETMHF-";
          break;
        case Muon:
          prefix = "PRECISION-MU-ETMHF-";
          break;
        default:
          TM_FATAL_ERROR("unsupported type: " << o2.getType());
      }
      break;

    default:
      TM_FATAL_ERROR("unsupported type: " << o1.getType());
  }
}


void
esTriggerMenuHandle::setFunctionCuts()
{
  std::map<std::string, unsigned int> dictionary;

  for (std::map<std::string, tmeventsetup::esScale>::const_iterator cit = std::begin(scale_map_);
       cit != std::end(scale_map_); cit++)
  {
    switch (cit->second.getScaleType())
    {
      case DeltaPrecision:
      case MassPrecision:
      case TwoBodyPtPrecision:
      case OvRmDeltaPrecision:
        break;

      default:
        continue;
    }

    std::pair<std::map<std::string, unsigned int>::iterator, bool> rc;
    rc = dictionary.insert(std::make_pair(cit->first, cit->second.getNbits()));
    if (not rc.second)
    {
      TM_FATAL_ERROR("insertion failure: " << TM_QUOTE(cit->first));
    }
  }

  if (dictionary.empty()) return; // hm, why bail out?


  for (const auto& condition: condition_map_)
  {
    const auto& objects = condition.second.getObjects();
    const auto& cuts = condition.second.getCuts();

    for (size_t ii = 0; ii < cuts.size(); ii++)
    {
      tmeventsetup::esCut& cut = const_cast<tmeventsetup::esCut&>(cuts.at(ii)); // TODO
      const int type = cut.getCutType();
      switch (type)
      {
        case DeltaEta:
        case DeltaPhi:
        case DeltaR:
        case Mass:
        case MassUpt:
        case MassDeltaR:
        case TwoBodyPt:
        case OvRmDeltaEta:
        case OvRmDeltaPhi:
        case OvRmDeltaR:
          break;
        default:
          continue;
      }

      std::string key;
      setPrefix4Precision(objects, key);

      double minimum = cut.getMinimumValue();
      double maximum = cut.getMaximumValue();
      unsigned int precision = 0;

      if (type == Mass or type == MassUpt or type == MassDeltaR)
      {
        key += PRECISION_MASS;
        if (not dictionary.count(key))
        {
          TM_FATAL_ERROR("missing precision scale set: " << TM_QUOTE(key));
        }
        precision = dictionary.at(key);
        // Note: calculate M^2/2 for mass
        const double scale = tmutil::pow10(precision);
        minimum = std::floor(minimum * minimum * 0.5 * scale) / scale;
        maximum = std::ceil(maximum * maximum * 0.5 * scale) / scale;
      }
      else if (type == DeltaR)
      {
        key += PRECISION_DELTA;
        if (not dictionary.count(key))
        {
          TM_FATAL_ERROR("missing precision scale set: " << TM_QUOTE(key));
        }
        precision = dictionary.at(key);
        // Note: calculate dR^2 for dR
        const double scale = tmutil::pow10(precision);
        minimum = std::floor(minimum * minimum * scale) / scale;
        maximum = std::ceil(maximum * maximum * scale) / scale;
      }
      else if (type == OvRmDeltaR)
      {
        key += PRECISION_OVRM_DELTA;
        if (not dictionary.count(key))
        {
          TM_FATAL_ERROR("missing precision scale set: " << TM_QUOTE(key));
        }
        precision = dictionary.at(key);
        // Note: calculate dR^2 for dR
        const double scale = tmutil::pow10(precision);
        minimum = std::floor(minimum * minimum * scale) / scale;
        maximum = std::ceil(maximum * maximum * scale) / scale;
      }
      else if (type == TwoBodyPt)
      {
        key += PRECISION_TBPT;
        if (not dictionary.count(key))
        {
          TM_FATAL_ERROR("missing precision scale set: " << TM_QUOTE(key));
        }
        precision = dictionary.at(key);
        const double scale = tmutil::pow10(precision);
        minimum = std::floor(minimum * minimum * scale) / scale;
        // TODO: actually this is not used for two body pt threshold
        maximum = std::ceil(maximum * maximum * scale) / scale;
      }
      else if (type == DeltaEta or type == DeltaPhi)
      {
        key += PRECISION_DELTA;
        if (not dictionary.count(key))
        {
          TM_FATAL_ERROR("missing precision scale set: " << TM_QUOTE(key));
        }
        precision = dictionary.at(key);
        const double scale = tmutil::pow10(precision);
        minimum = std::floor(minimum * scale) / scale;
        maximum = std::ceil(maximum * scale) / scale;
      }
      else if (type == OvRmDeltaEta or type == OvRmDeltaPhi)
      {
        key += PRECISION_OVRM_DELTA;
        if (not dictionary.count(key))
        {
          TM_FATAL_ERROR("missing precision scale set: " << TM_QUOTE(key));
        }
        precision = dictionary.at(key);
        const double scale = tmutil::pow10(precision);
        minimum = std::floor(minimum * scale) / scale;
        maximum = std::ceil(maximum * scale) / scale;
      }
      else
      {
        TM_FATAL_ERROR("unsupported cut " << TM_QUOTE(cut.getName()) << " of type: " << type);
      }

      cut.setMinimumValue(minimum);
      cut.setMaximumValue(maximum);
      cut.setPrecision(precision); // TODO set both minimum/maximum index
    } // for cut
  } // for condition
}

} // namespace tmeventsetup
