#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE tmEventSetup

#include "tmEventSetup/tmEventSetup.hh"

#include "tmEventSetup/esTriggerMenuHandle.hh"
#include "tmEventSetup/esTriggerMenu.hh"

#include "tmEventSetup/esAlgorithmHandle.hh"
#include "tmEventSetup/esAlgorithm.hh"

#include "tmEventSetup/esCutHandle.hh"
#include "tmEventSetup/esCut.hh"

#include "tmEventSetup/esScaleHandle.hh"
#include "tmEventSetup/esScale.hh"

#include "tmUtil/tmUtil.hh"

#include <boost/test/unit_test.hpp>

#include <algorithm>
#include <string>
#include <vector>

#include <cmath>

namespace tmeventsetup
{
  extern long long toFixedPoint(double, const size_t);
}

BOOST_AUTO_TEST_SUITE(esAlgorithm)

BOOST_AUTO_TEST_CASE(esAlgorithmHandle_init)
{
  const tmtable::Row row{
    {"name", "L1_Unnamed"},
    {"index", "42"},
    {"module_id", "0"},
    {"module_index" ,"1"},
    {"expression", "mass_inv{MU4,MU2}[MASS_Z] OR TAU42"}
  };

  tmeventsetup::esAlgorithmHandle handle;
  handle.init(row);

  tmeventsetup::esAlgorithm* algorithm = dynamic_cast<tmeventsetup::esAlgorithm*>(&handle);

  BOOST_CHECK_EQUAL("L1_Unnamed", algorithm->getName());
  BOOST_CHECK_EQUAL(42, algorithm->getIndex());
  BOOST_CHECK_EQUAL(0, algorithm->getModuleId());
  BOOST_CHECK_EQUAL(1, algorithm->getModuleIndex());
  BOOST_CHECK_EQUAL("mass_inv{MU4,MU2}[MASS_Z] OR TAU42", algorithm->getExpression());
  BOOST_CHECK_EQUAL("", algorithm->getExpressionInCondition());
}

BOOST_AUTO_TEST_CASE(esAlgorithmHandle_setRpnVector)
{
  const tmtable::Row row{
    {"name", "L1_setRpnVector"},
    {"index", "42"},
    {"module_id", "0"},
    {"module_index", "1"},
    {"expression", "MU7 AND TAU3"}
  };

  const std::map<std::string, std::string> conditionMap{
    {"MU7", "SingleMu_4"},
    {"TAU3", "SingleTau_5"}
  };

  tmeventsetup::esAlgorithmHandle handle;
  handle.init(row);
  handle.setExpressionInCondition(conditionMap);

  tmeventsetup::esAlgorithm* algorithm = dynamic_cast<tmeventsetup::esAlgorithm*>(&handle);

  BOOST_CHECK_EQUAL("L1_setRpnVector", algorithm->getName());
  BOOST_CHECK_EQUAL(42, algorithm->getIndex());
  BOOST_CHECK_EQUAL(0, algorithm->getModuleId());
  BOOST_CHECK_EQUAL(1, algorithm->getModuleIndex());
  BOOST_CHECK_EQUAL("MU7 AND TAU3", algorithm->getExpression());
  BOOST_CHECK_EQUAL("SingleMu_4 AND SingleTau_5", algorithm->getExpressionInCondition());
}

BOOST_AUTO_TEST_CASE(esAlgorithmHandle_setExpressionInCondition)
{
  const tmtable::Row row{
    {"name", "L1_setExpressionInCondition"},
    {"index", "42"},
    {"module_id", "0"},
    {"module_index", "1"},
    {"expression", "mass_inv{MU4,MU2}[MASS_Z] OR TAU42 AND EXT_SIG_0"}
  };

  const std::map<std::string, std::string> conditionMap{
    {"mass_inv{MU4,MU2}[MASS_Z]", "mass_123"},
    {"TAU42", "single_tau_124"},
    {"EXT_SIG_0", "ext_signal_125"}
  };

  tmeventsetup::esAlgorithmHandle handle;
  handle.init(row);
  handle.setExpressionInCondition(conditionMap);

  // Isn't this all a quite a big rubish?!
  tmeventsetup::esTriggerMenuHandle triggerMenuHandle;
  handle.setRpnVector(triggerMenuHandle.parse(handle.getExpressionInCondition()));

  tmeventsetup::esAlgorithm* algorithm = dynamic_cast<tmeventsetup::esAlgorithm*>(&handle);

  BOOST_CHECK_EQUAL("L1_setExpressionInCondition", algorithm->getName());
  BOOST_CHECK_EQUAL(42, algorithm->getIndex());
  BOOST_CHECK_EQUAL(0, algorithm->getModuleId());
  BOOST_CHECK_EQUAL(1, algorithm->getModuleIndex());
  BOOST_CHECK_EQUAL("mass_inv{MU4,MU2}[MASS_Z] OR TAU42 AND EXT_SIG_0", algorithm->getExpression());
  BOOST_CHECK_EQUAL("mass_123 OR single_tau_124 AND ext_signal_125", algorithm->getExpressionInCondition());

  const std::vector<std::string> rpn_vector_ref{
    "mass_123",
    "single_tau_124",
    "OR",
    "ext_signal_125",
    "AND"
  };
  std::vector<std::string> rpn_vector = algorithm->getRpnVector();

  BOOST_CHECK_EQUAL_COLLECTIONS(rpn_vector.begin(), rpn_vector.end(),
                                rpn_vector_ref.begin(), rpn_vector_ref.end());
}

BOOST_AUTO_TEST_SUITE_END()

BOOST_AUTO_TEST_SUITE(esCut)

BOOST_AUTO_TEST_CASE(esCutHandle_init)
{
  const tmtable::Row row{
    {"name", "MU-QLTY_OPEN"},
    {"object", "MU"},
    {"type", "MU-QLTY"}, // Note: type is overwritten in esObjectHandle!
    {"minimum" ,"+0.0000000000000000E+00"},
    {"maximum" ,"+0.0000000000000000E+00"},
    {"data", ""}
  };

  tmeventsetup::esCutHandle handle(row);
  // handle.init(row); not implemented?

  tmeventsetup::esCut* cut = dynamic_cast<tmeventsetup::esCut*>(&handle);

  BOOST_CHECK_EQUAL("MU-QLTY_OPEN", cut->getName());
  BOOST_CHECK_EQUAL(tmeventsetup::Muon, cut->getObjectType());
  BOOST_CHECK_EQUAL(tmeventsetup::Quality, cut->getCutType());
  BOOST_CHECK_EQUAL(0, cut->getMinimumValue());
  BOOST_CHECK_EQUAL(0, cut->getMaximumValue());
  BOOST_CHECK_EQUAL("0", cut->getData());
  BOOST_CHECK_EQUAL("MU-QLTY", cut->getKey());
  BOOST_CHECK_EQUAL(std::numeric_limits<unsigned int>::max(), cut->getPrecision());
}

BOOST_AUTO_TEST_CASE(esCutHandle_init_jet_disp)
{
  const tmtable::Row row{
    {"name", "JET-DISP_LLP"},
    {"object", "JET"},
    {"type", "JET-DISP"}, // Note: type is overwritten in esObjectHandle!
    {"minimum" ,"+0.0000000000000000E+00"},
    {"maximum" ,"+0.0000000000000000E+00"},
    {"data", "1"}
  };

  tmeventsetup::esCutHandle handle(row);
  // handle.init(row); not implemented?

  tmeventsetup::esCut* cut = dynamic_cast<tmeventsetup::esCut*>(&handle);

  BOOST_CHECK_EQUAL("JET-DISP_LLP", cut->getName());
  BOOST_CHECK_EQUAL(tmeventsetup::Jet, cut->getObjectType());
  BOOST_CHECK_EQUAL(tmeventsetup::Displaced, cut->getCutType());
  BOOST_CHECK_EQUAL(0, cut->getMinimumValue());
  BOOST_CHECK_EQUAL(0, cut->getMaximumValue());
  BOOST_CHECK_EQUAL("1", cut->getData());
  BOOST_CHECK_EQUAL("JET-DISP", cut->getKey());
  BOOST_CHECK_EQUAL(std::numeric_limits<unsigned int>::max(), cut->getPrecision());
}

BOOST_AUTO_TEST_SUITE_END()

BOOST_AUTO_TEST_SUITE(esScale)

BOOST_AUTO_TEST_CASE(esScaleHandle_et)
{
  const tmtable::Row row{
    {"object", "JET"},
    {"type", "ET"},
    {"minimum", "+0.0000000000000000E+00"},
    {"maximum" ,"+127.0000000000000000E+00"},
    {"step", "+1.0000000000000000E+00"},
    {"n_bits", "7"}
  };

  tmeventsetup::esScaleHandle handle(row);

  tmeventsetup::esScale* scale = dynamic_cast<tmeventsetup::esScale*>(&handle);

  BOOST_CHECK_EQUAL("JET-ET", scale->getName());
  BOOST_CHECK_EQUAL(tmeventsetup::Jet, scale->getObjectType());
  BOOST_CHECK_EQUAL(tmeventsetup::EtScale, scale->getScaleType());
  BOOST_CHECK_EQUAL(0.0, scale->getMinimum());
  BOOST_CHECK_EQUAL(127.0, scale->getMaximum());
  BOOST_CHECK_EQUAL(1.0, scale->getStep());
  BOOST_CHECK_EQUAL(7, scale->getNbits());
}

BOOST_AUTO_TEST_CASE(esScaleHandle_upt)
{
  const tmtable::Row row{
    {"object", "MU"},
    {"type", "UPT"},
    {"minimum", "+0.0000000000000000E+00"},
    {"maximum" ,"+255.0000000000000000E+00"},
    {"step", "+1.0000000000000000E+00"},
    {"n_bits", "8"}
  };

  tmeventsetup::esScaleHandle handle(row);

  tmeventsetup::esScale* scale = dynamic_cast<tmeventsetup::esScale*>(&handle);

  BOOST_CHECK_EQUAL("MU-UPT", scale->getName());
  BOOST_CHECK_EQUAL(tmeventsetup::Muon, scale->getObjectType());
  BOOST_CHECK_EQUAL(tmeventsetup::UnconstrainedPtScale, scale->getScaleType());
  BOOST_CHECK_EQUAL(0.0, scale->getMinimum());
  BOOST_CHECK_EQUAL(255.0, scale->getMaximum());
  BOOST_CHECK_EQUAL(1.0, scale->getStep());
  BOOST_CHECK_EQUAL(8, scale->getNbits());
}

BOOST_AUTO_TEST_CASE(esScaleHandle_precision_delta)
{
  const tmtable::Row row{
    {"object", "PRECISION"},
    {"type", "TAU-MU-Delta"},
    {"minimum", "+0.0000000000000000E+00"},
    {"maximum" ,"+0.0000000000000000E+00"},
    {"step", "+0.0000000000000000E+00"},
    {"n_bits", "4"}
  };

  tmeventsetup::esScaleHandle handle(row);

  tmeventsetup::esScale* scale = dynamic_cast<tmeventsetup::esScale*>(&handle);

  BOOST_CHECK_EQUAL("PRECISION-TAU-MU-Delta", scale->getName());
  BOOST_CHECK_EQUAL(tmeventsetup::Precision, scale->getObjectType());
  BOOST_CHECK_EQUAL(tmeventsetup::DeltaPrecision, scale->getScaleType());
  BOOST_CHECK_EQUAL(0.0, scale->getMinimum());
  BOOST_CHECK_EQUAL(0.0, scale->getMaximum());
  BOOST_CHECK_EQUAL(0.0, scale->getStep());
  BOOST_CHECK_EQUAL(4, scale->getNbits());
}

BOOST_AUTO_TEST_CASE(esScaleHandle_precision_delta_overlap_removal)
{
  const tmtable::Row row{
    {"object", "PRECISION"},
    {"type", "MU-MU-DeltaOverlapRemoval"},
    {"minimum", "+0.0000000000000000E+00"},
    {"maximum" ,"+0.0000000000000000E+00"},
    {"step", "+0.0000000000000000E+00"},
    {"n_bits", "9"}
  };

  tmeventsetup::esScaleHandle handle(row);

  tmeventsetup::esScale* scale = dynamic_cast<tmeventsetup::esScale*>(&handle);

  BOOST_CHECK_EQUAL("PRECISION-MU-MU-DeltaOverlapRemoval", scale->getName());
  BOOST_CHECK_EQUAL(tmeventsetup::Precision, scale->getObjectType());
  BOOST_CHECK_EQUAL(tmeventsetup::OvRmDeltaPrecision, scale->getScaleType());
  BOOST_CHECK_EQUAL(0.0, scale->getMinimum());
  BOOST_CHECK_EQUAL(0.0, scale->getMaximum());
  BOOST_CHECK_EQUAL(0.0, scale->getStep());
  BOOST_CHECK_EQUAL(9, scale->getNbits());
}

BOOST_AUTO_TEST_CASE(esScaleHandle_precision_mass)
{
  const tmtable::Row row{
    {"object", "PRECISION"},
    {"type", "EG-TAU-Mass"},
    {"minimum", "+0.0000000000000000E+00"},
    {"maximum" ,"+0.0000000000000000E+00"},
    {"step", "+0.0000000000000000E+00"},
    {"n_bits", "1"}
  };

  tmeventsetup::esScaleHandle handle(row);

  tmeventsetup::esScale* scale = dynamic_cast<tmeventsetup::esScale*>(&handle);

  BOOST_CHECK_EQUAL("PRECISION-EG-TAU-Mass", scale->getName());
  BOOST_CHECK_EQUAL(tmeventsetup::Precision, scale->getObjectType());
  BOOST_CHECK_EQUAL(tmeventsetup::MassPrecision, scale->getScaleType());
  BOOST_CHECK_EQUAL(0.0, scale->getMinimum());
  BOOST_CHECK_EQUAL(0.0, scale->getMaximum());
  BOOST_CHECK_EQUAL(0.0, scale->getStep());
  BOOST_CHECK_EQUAL(1, scale->getNbits());
}

BOOST_AUTO_TEST_SUITE_END()

BOOST_AUTO_TEST_SUITE(esTriggerMenu)

BOOST_AUTO_TEST_CASE(esTriggerMenu_attributes)
{
  // References
  const std::string name = "L1Menu_Foobar";
  const std::string version = "1.42";
  const std::string comment = "NO-body expects the Spanish Inquisition!";
  const std::string datetime = "2016-12-25T18:00:00+0000";
  const std::string firmwareUuid = "f06d0803-1852-4ac8-9f79-01299371a28b";
  const std::string scaleSetName = "spam";
  const size_t nModules = 4;

  // Create menu container
  tmeventsetup::esTriggerMenu menu;

  // Populate attribuites
  menu.setName(name);
  menu.setVersion(version);
  menu.setComment(comment);
  menu.setDatetime(datetime);
  menu.setFirmwareUuid(firmwareUuid);
  menu.setScaleSetName(scaleSetName);
  menu.setNmodules(nModules);

  // Check attributes
  BOOST_CHECK_EQUAL(name, menu.getName());
  BOOST_CHECK_EQUAL(version, menu.getVersion());
  BOOST_CHECK_EQUAL(comment, menu.getComment());
  BOOST_CHECK_EQUAL(datetime, menu.getDatetime());
  BOOST_CHECK_EQUAL(firmwareUuid, menu.getFirmwareUuid());
  BOOST_CHECK_EQUAL(scaleSetName, menu.getScaleSetName());
  BOOST_CHECK_EQUAL(nModules, menu.getNmodules());
}

BOOST_AUTO_TEST_SUITE_END()

BOOST_AUTO_TEST_SUITE(esTriggerMenuHandle)

BOOST_AUTO_TEST_SUITE_END()

BOOST_AUTO_TEST_SUITE(tmEventSetup)

BOOST_AUTO_TEST_CASE(tmEventSetup_getValue)
{
  const std::map<std::string, std::string> data{{"foo", "42"}};

  BOOST_CHECK_EQUAL(tmeventsetup::getValue(data, ""), "");
  BOOST_CHECK_EQUAL(tmeventsetup::getValue(data, "foo"), "42");
  BOOST_CHECK_EQUAL(tmeventsetup::getValue(data, "bar"), "");
}

BOOST_AUTO_TEST_CASE(tmEventSetup_getMmHashN)
{
  unsigned int seed = 3735927486; // fixed inside getMmHashN

  std::string phrase = "NO-body expects the Spanish Inquisition!";

  BOOST_CHECK_EQUAL(
    tmeventsetup::getMmHashN(phrase),
    tmutil::MurmurHashNeutral2(phrase.c_str(), phrase.size(), seed)
  );
}

BOOST_AUTO_TEST_CASE(tmEventSetup_getExternalName)
{
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName(""), "");
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName("EXT_"), "");
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName("+1"), "");
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName("-1"), "");
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName("EXT_A+B+C"), "A");
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName("EXT_FOO"), "FOO");
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName("EXT_FOO+0"), "FOO");
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName("EXT_FOO-0"), "FOO");
  BOOST_CHECK_EQUAL(tmeventsetup::getExternalName("EXT_FOO_BAR+2"), "FOO_BAR");
}

BOOST_AUTO_TEST_CASE(tmEventSetup_getObjectCombination)
{
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Muon, tmeventsetup::Muon), tmeventsetup::MuonMuonCombination);

  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Muon, tmeventsetup::ETM), tmeventsetup::MuonEsumCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Muon, tmeventsetup::HTM), tmeventsetup::MuonEsumCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Muon, tmeventsetup::ETMHF), tmeventsetup::MuonEsumCombination);

  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Muon, tmeventsetup::Egamma), tmeventsetup::CaloMuonCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Muon, tmeventsetup::Tau), tmeventsetup::CaloMuonCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Muon, tmeventsetup::Jet), tmeventsetup::CaloMuonCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Muon, tmeventsetup::Egamma), tmeventsetup::CaloMuonCombination);

  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Egamma, tmeventsetup::Egamma), tmeventsetup::CaloCaloCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Egamma, tmeventsetup::Tau), tmeventsetup::CaloCaloCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Egamma, tmeventsetup::Jet), tmeventsetup::CaloCaloCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Tau, tmeventsetup::Egamma), tmeventsetup::CaloCaloCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Tau, tmeventsetup::Tau), tmeventsetup::CaloCaloCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Tau, tmeventsetup::Jet), tmeventsetup::CaloCaloCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Jet, tmeventsetup::Egamma), tmeventsetup::CaloCaloCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Jet, tmeventsetup::Tau), tmeventsetup::CaloCaloCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Jet, tmeventsetup::Jet), tmeventsetup::CaloCaloCombination);

  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Egamma, tmeventsetup::ETM), tmeventsetup::CaloEsumCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Egamma, tmeventsetup::HTM), tmeventsetup::CaloEsumCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Egamma, tmeventsetup::ETMHF), tmeventsetup::CaloEsumCombination);

  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Tau, tmeventsetup::ETM), tmeventsetup::CaloEsumCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Tau, tmeventsetup::HTM), tmeventsetup::CaloEsumCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Tau, tmeventsetup::ETMHF), tmeventsetup::CaloEsumCombination);

  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Jet, tmeventsetup::ETM), tmeventsetup::CaloEsumCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Jet, tmeventsetup::HTM), tmeventsetup::CaloEsumCombination);
  BOOST_CHECK_EQUAL(tmeventsetup::getObjectCombination(tmeventsetup::Jet, tmeventsetup::ETMHF), tmeventsetup::CaloEsumCombination);
}

BOOST_AUTO_TEST_CASE(tmEventSetup_getPrecisions)
{
  // TODO
}

BOOST_AUTO_TEST_CASE(tmEventSetup_toFixedPoint)
{
  BOOST_CHECK_EQUAL(0, tmeventsetup::toFixedPoint(0.123, 0));
  BOOST_CHECK_EQUAL(1, tmeventsetup::toFixedPoint(0.123, 1));
  BOOST_CHECK_EQUAL(12, tmeventsetup::toFixedPoint(0.123, 2));
  BOOST_CHECK_EQUAL(123, tmeventsetup::toFixedPoint(0.123, 3));

  BOOST_CHECK_EQUAL(123, tmeventsetup::toFixedPoint(123.456, 0));
  BOOST_CHECK_EQUAL(1235, tmeventsetup::toFixedPoint(123.456, 1));
  BOOST_CHECK_EQUAL(12346, tmeventsetup::toFixedPoint(123.456, 2));
  BOOST_CHECK_EQUAL(123456, tmeventsetup::toFixedPoint(123.456, 3));

  // Check rounding
  BOOST_CHECK_EQUAL(14, tmeventsetup::toFixedPoint(1.449, 1));
  BOOST_CHECK_EQUAL(15, tmeventsetup::toFixedPoint(1.499, 1));
  BOOST_CHECK_EQUAL(15, tmeventsetup::toFixedPoint(1.500, 1));
}

BOOST_AUTO_TEST_CASE(tmEventSetup_setLut)
{
  size_t precision = 1;

  const std::vector<double> values {0.123, 1.234, 2.345, 3.456 };

  // Reference set with precision of 1
  const std::vector<long long> reference{1, 12, 23, 35};

  std::vector<long long> results;
  tmeventsetup::setLut(results, values, precision);

  BOOST_CHECK_EQUAL_COLLECTIONS(results.begin(), results.end(),
                                reference.begin(), reference.end());
}

BOOST_AUTO_TEST_CASE(tmEventSetup_applySin)
{
  std::vector<double> values{0.123, 1.234, 2.345, 3.456, 4.567, 5.678};

  std::vector<double> reference(values.size());
  std::transform(values.begin(), values.end(), reference.begin(), ::sin);

  tmeventsetup::applySin(values, values.size());

  BOOST_CHECK_EQUAL_COLLECTIONS(values.begin(), values.end(),
                                reference.begin(), reference.end());
}

BOOST_AUTO_TEST_CASE(tmEventSetup_applyCos)
{
  std::vector<double> values{0.123, 1.234, 2.345, 3.456, 4.567, 5.678};

  std::vector<double> reference(values.size());
  std::transform(values.begin(), values.end(), reference.begin(), ::cos);

  tmeventsetup::applyCos(values, values.size());

  BOOST_CHECK_EQUAL_COLLECTIONS(values.begin(), values.end(),
                                reference.begin(), reference.end());
}

BOOST_AUTO_TEST_CASE(tmEventSetup_applyCosh)
{
  std::vector<double> values{0.123, 1.234, 2.345, 3.456, 4.567, 5.678};

  std::vector<double> reference(values.size());
  std::transform(values.begin(), values.end(), reference.begin(), ::cosh);

  tmeventsetup::applyCosh(values, values.size());

  BOOST_CHECK_EQUAL_COLLECTIONS(values.begin(), values.end(),
                                reference.begin(), reference.end());
}

BOOST_AUTO_TEST_SUITE_END()
